from .mapping import add_stamen_basemap
from .mapping import add_north_arrow
from .mapping import add_scale_bar
from .mapping import ax_ticks
from .mapping import latlon_to_utm_epsg
from .mapping import set_min_axes
from .mapping import set_figure_size
from .mapping import add_POI_marker
from .mapping import stamen_zoom_level
from .epsg_defs import webmerc, wgs84
from .webmap import BindColormap